"""
This file contains the functions needed in order to generate an automatic report.
"""

#Imports
import os
from reportlab.lib import colors
from reportlab.lib.pagesizes import letter
from reportlab.lib.units import inch
from reportlab.platypus import SimpleDocTemplate, Table, Paragraph, Spacer, Image
from reportlab.lib.styles import getSampleStyleSheet

from .Basic_functions import make_solution
from .Solutions_Solubility import plot_graph, data4graph
from .PH_Approximation_0 import *
from .pH_graph_0 import *

# Auxiliar functions
def merge_dicts(dict1, dict2, ions_of_interest = "all"):
    # Get the union of keys from both dictionaries
    all_ions = set(dict1.keys()) | set(dict2.keys())
    
    # Create lists to store ion names, values from dict1, and values from dict2
    ion_names = []
    values_dict1 = []
    values_dict2 = []
    
    # Iterate over ion names
    for ion in all_ions:
        # Append ion name to ion_names list
        if ion in ions_of_interest or ions_of_interest == "all":
            ion_names.append(ion)
        
        # Append value from dict1 or None if key does not exist
        if ion in dict1 and ion in ion_names:
            values_dict1.append(round(dict1[ion],4))
        elif ion in ion_names:
            values_dict1.append(0)
        
        # Append value from dict2 or None if key does not exist
        if ion in dict2 and ion in ion_names:
            values_dict2.append(round(dict2[ion],4))
        elif ion in ion_names:
            values_dict2.append(0)
    
    return [ion_names, values_dict1, values_dict2]

#---------- Main function ---------------------------------

def generate_report(
    plant_name: str,
    required_nutriments: dict,
    growth_time: float,
    ions_of_interest: list,
    solution_composition: dict,
    solution_volume: float,
    forbidden_ions: list = [],

) -> None:
    """
    Generates a simulation report in PDF format based on the provided parameters.

    Args:
        plant_name (str): The name of the plant for which the hydroponic solution simulation is conducted.
        required_nutriments (dict): A dictionary containing the required nutriments (ions) for the specified plant for a full growth [g].
        solution_composition (dict): A dictionary representing the composition of the hydroponic solution [g/L] for each ion.
        solution_volume (float): The volume of the hydroponic solution [L].
        ions_of_interest (list): A list of ions for which the concentration is of interest.
        growth_time (float): The duration of the growth of the plant [days].
        forbidden_ions (list, default): A list of ions that are forbidden in the solution.

    Returns:
        None: a pdf is created in the folder of this file
    """
    print("Generating report...")
    # PageTemplate of the document
    doc = SimpleDocTemplate("hydroponic_report.pdf", pagesize=letter)
    story = []
    interest_ions = ", ".join(ions_of_interest)
    nono_ions = ", ".join(forbidden_ions)
    
    # Title
    title_style = getSampleStyleSheet()["Title"]
    title = Paragraph("Hydroponic Farming Simulation Report", title_style)
    story.append(title)
    story.append(Spacer(1, 12))

    # 1 .Simulation conditions ---------------------------------------------------------
    subtitle_style = getSampleStyleSheet()['Heading2']
    subtitle = Paragraph(" 1. Simulation Conditions", subtitle_style)
    story.append(subtitle)
    story.append(Spacer(1, 6))
    
    paragraph_style = getSampleStyleSheet()["Normal"]
    conditions_text = f"This report summarizes the simulation of the growth of a {plant_name} in a hydroponic solution.<br/>"\
    f"The volume of the solution used in the simulation is {solution_volume} L. "\
    f"The simulation was done for a duration of {growth_time} days and the followig ions were analysed: {interest_ions}.<br/> <br/>"\
    f"The following assumptions were made for the simulation:<br/>"\
        f" - The plant consumes a constant amount of nutriments each day <br/>"\
        f" - The temperature is assumed to be constant at 25 °C and the pressure at 1 atm. <br/>"\
        f" - The effect of the sun exposure and light conditions was neglected. <br/> <br/>"\
    f"The following elements were simulated and analysed:<br/>"\
        f" - The amount of salts to add to the solution. <br/>"\
        f" - The evolution of the concentration of the ions of interest in the solution. <br/>"\
        f" - The pH of the solution. <br/>"\
    f"<br/> <br/>"\
    f" The table below shows the provided optimal concentrations of the ions in the hydroponic solution and "\
    f"the nutritional needs of the {plant_name} plant (fully grown):"
    story.append(Paragraph(conditions_text, paragraph_style))
    story.append(Spacer(1, 12))
    
    #Table 
    data = merge_dicts(solution_composition, required_nutriments, ions_of_interest)
    headers = ["Ions",'Hydroponic solution [g/L]', f'{plant_name} nutriments requirements[g]']
    table_data = [[elem] + sublist for elem, sublist in zip(headers, data)] 
    # Transpose the data (swap rows and columns)
    table_data = list(map(list, zip(*table_data)))
    table = Table(table_data)
    table.setStyle([('BACKGROUND', (0, 0), (-1, 0), colors.grey),
                    ('TEXTCOLOR', (0, 0), (-1, 0), colors.whitesmoke),
                    ('ALIGN', (0, 0), (-1, -1), 'CENTER'),
                    ('FONTNAME', (0, 0), (-1, 0), 'Helvetica-Bold'),
                    ('BOTTOMPADDING', (0, 0), (-1, 0), 12)])
    # Calculate required size for the table
    table.wrap(doc.width, doc.height)
    
    # Add the table to the story
    story.append(table)
    story.append(Spacer(1, 12))
    print("Collecting the data ...")
    # 2. Preparation of the solution ---------------------------------------------------------
    subtitle = Paragraph(" 2. Preparation of the solution", subtitle_style)
    story.append(subtitle)
    story.append(Spacer(1, 6))
    
    preparation_text = f"The table below shows the mass of salts to add the the {solution_volume} L solution" \
        f" to obtain the desired concentrations of the ions for the {plant_name} plant.<br/>"\
        f" The following restrictions were given for the preparation of the solution:<br/>"\
        f" - Forbidden ions: {nono_ions} <br/>"\
        f" - All the salts must be soluble. <br/>"\
        f" - The salts come from a list of commercially available salts. <br/>"
        
    story.append(Paragraph(preparation_text, paragraph_style))
    story.append(Spacer(1, 12))
    
    #Add table: "salts to add in the solution"
    salt_composition = make_solution(solution_composition, forbidden_ions ,solution_volume)
    headers = ["Salts"]+list(salt_composition.keys())
    data = list(salt_composition.values())
    data = [round(float(elem), 4) for elem in data]
    data = [f"Mass [g] for {solution_volume} L solution"]+data
    table_data = [[elem] + [sublist] for elem, sublist in zip(headers, data)]
    table = Table(table_data)
    table.setStyle([('BACKGROUND', (0, 0), (-1, 0), colors.grey),
                    ('TEXTCOLOR', (0, 0), (-1, 0), colors.whitesmoke),
                    ('ALIGN', (0, 0), (-1, -1), 'CENTER'),
                    ('FONTNAME', (0, 0), (-1, 0), 'Helvetica-Bold'),
                    ('BOTTOMPADDING', (0, 0), (-1, 0), 12)])
    # Calculate required size for the table
    table.wrap(doc.width, doc.height)
    story.append(table)
    story.append(Spacer(1, 12))
    
    
    # 3. Ion concentration elvolution ---------------------------------------------------------
    subtitle = Paragraph(" 3. Evolution of the concentration of ions", subtitle_style)
    story.append(subtitle)
    story.append(Spacer(1, 6))
    
    ion_figure_text = f"The Figure below shows the concentration of the ions of interest ({interest_ions}) " \
        f"as a function of the days of growth of the plant."
    story.append(Paragraph(ion_figure_text, paragraph_style))
    story.append(Spacer(1, 12))
    
    # Make the graph
    plot_graph(solution_composition, "ion", required_nutriments, growth_time, solution_volume, ions_of_interest)
    
    # Get the path of the graph
    current_dir = os.path.dirname(os.path.realpath(__file__))
    file_name = "graph_"
    for ion in ions_of_interest:
        file_name += str(ion) + "_"
    file_name += ".png"
    file_path = os.path.join(current_dir, file_name)

    # Add the image to the story and center it
    image = Image(file_path, width=600, height=300)
    image.hAlign = 'CENTER'
    story.append(image)
    story.append(Spacer(1, 12))
    
    print("Simulating the plant growth...")
    # 4. pH of the solution ---------------------------------------------------------
    subtitle = Paragraph(" 4. pH of the solution", subtitle_style)
    story.append(subtitle)
    story.append(Spacer(1, 6))
    
    pH_text = f"The Figure below shows the pH of the solution as a function of the growth of the plant."\
        f"The pH limits are shown in the graph."
    
    story.append(Paragraph(pH_text, paragraph_style))
    story.append(Spacer(1, 12))
    
    #Generate the data
    solution_pH = data4graph(solution_composition,solution_volume,required_nutriments,growth_time)[1]
    temp = int(25)
    pH_part_of_report_generation(solution_pH, temp, plant_name) #create images
    
    # Get the path of the pH_graph
    current_dir = os.path.dirname(os.path.realpath(__file__))
    file_name = "graph_pH.png"
    file_path = os.path.join(current_dir, file_name)
    # Add the image to the story and center it
    image = Image(file_path, width=600, height=300)
    image.hAlign = 'CENTER'
    story.append(image)
    story.append(Spacer(1, 12))
    
    # Get the path of the pH_table
    current_dir = os.path.dirname(os.path.realpath(__file__))
    file_name = "Table_pH.png"
    file_path = os.path.join(current_dir, file_name)
    # Add the image to the story and center it
    image = Image(file_path, width=300, height=300)
    image.hAlign = 'CENTER'
    story.append(image)
    story.append(Spacer(1, 12))
    
    # Build the PDF
    doc.build(story)
    print("Report generated successfully!")
    



from sympy import symbols, Eq, solve
import math
import random

def pH_approximation (concentration_of_ions_in_solution:dict, temperature:float)->float:
    """
    This function calculates the pH of a solution given the concentrations of ions in the solution and the temperature of the solution.

    Parameters:
    concentration_of_ions_in_solution (dict): A dictionary containing the ions(str) as keys and their concentrations(float) in the solution as values.
    temperature (float): The temperature of the solution in degrees Celsius.

    Returns:
    pH (float): The pH of the solution. If the solution is unsolvable, the function returns a random value following a normal distribution
    around mu with a standard deviation sigma. 
    """
    try:
        #Using find_acid so as to determine pKa values associated to some ions in the solution 
        Degree_of_deprotonation, pKa_identified_ions,highest_existing_charge_of_identified_ion=find_acid(concentration_of_ions_in_solution, pKa_values)
        #setting index for compounds possesing pKa values
        i=0
        #setting index for compounds possesing Ksp values (metals)
        l=0
        #setting variable of the concentration of protons in solution
        H_plus_concentration=symbols("H_plus_concentration")
        #setting variable of the concentration of OH- in solution 
        OH_minus_concentration=symbols("OH_minus_concentration")
        
        # Defines system of equations
        equations = []
        
        # Initialize a list to accumulate variables for charge balance calculations 
        accumulated_variables_for_charge_balance = []
        # Assigns the value of the ionization constant of water (Kw) at a specific temperature
        Kw=Value_of_Ionisation_Constant[temperature]
        # Defines the equation for the ionization constant of water
        Kw_equation=Eq(H_plus_concentration*OH_minus_concentration,Kw)
        equations.append(Kw_equation)
        #Adding OH_minus and H_plus to charge balance
        accumulated_variables_for_charge_balance.append(H_plus_concentration-OH_minus_concentration)
        #Initialise a list for final proton concentration
        accumulated_variables_for_final_proton_concentration=[]
        #Initialise a variable for final OH-
        accumulated_variables_for_final_OH_concentration=[]
        #Initial concentration of protons
        H_plus_zero=10**(-7)
        #Initial concentration of OH-
        OH_minus_zero=10**(-7)
        #Adding final concentration variables
        accumulated_variables_for_final_OH_concentration.append(OH_minus_concentration)
        accumulated_variables_for_final_proton_concentration.append(H_plus_concentration)
        for compound, compound_concentration in concentration_of_ions_in_solution.items():
            #If the concentration of the compound is 0, then the compound is not considered in the calculations
            if float(compound_concentration) == 0:
                continue
            """
            Below, it is checked as to whether or not the compound has a solubility product constant or an acid dissociation constant. The former allows one to define the impact that metal ions have on the pH of the solution
            One can see Ksp having an impact on the pH with the example that follows. We have an initial salt we place in a solution, i.e, CaSO4, which then dissociates into Ca^2+ and SO4^2-. 
            There is then a Ksp constant that depends on the concentration of Ca^2+, OH^-1 and Ca(OH)2 and a Ka value that depends on the concentration of SO4^2-, HSO4^1- and H2SO4 as well as H+
            The Ksp constant has an impact on the concentration of OH^-1 which in itself impact the pH of the solution.
            """
            if compound in Ksp_values:
                #Defines variables for the concentrations of the metal in a dissociated state or not, i.e, Ca(OH)^2- or Ca^2+
                exec(f"dissociated_metal_{l}_concentration]=symbols(dissociated_metal_{l}_concentration)")
                exec(f"undissociated_metal_compound_{l}_concentration=symbols(undissociated_metal_compound_{l}_concentration)")
                #Equation of dissociation equilibrium
                exec(f"equation_Ksp = Eq((dissociated_metal_compound_{l}_concentration*OH_minus_concentration**(compound_charge[compound])), Ksp_values[compound])")
                equations.append(equation_Ksp)
                # Accumulating variables for charge balance. 
                accumulated_variables_for_charge_balance.append(compound_charge[compound]*f'dissociated_metal_{l}_concentration')
                #Mass equation of metal/metal compound
                exec(f"mass_balance_metal = Eq(dissociated_metal_{l}_concentration + undissociated_metal_{l}_concentration, compound_concentration)")
                equations.append(mass_balance_metal)
                exec(f"accumulated_variables_for_final_OH_concentration.append(-compound_charge[compound]*undissociated_metal_compound_{l}_concentration)")
                l+=1
            else: 
                #Defines the constant that is the initial concentration of the compound 
                globals()[f'initial_compound_concentration_{i}']=compound_concentration 
                #Initialize a variable to accumulate the variables of the concentrations into one equation of mass balance
                accumulated_variables_for_mass_balance=[]
                #Determines the number of states a compound is in whose concentrations define one of the known Ka values of the compound, there could be more states in reality, but they are negligeable. 
                number_of_states_of_dissociation=len(pKa_identified_ions[compound])+1
                
                for s in range (number_of_states_of_dissociation): 
                    #derives the Ka_th value of a compound
                    if s!=number_of_states_of_dissociation-1:
                        exec(f'Ka_{s}_value_of_compound_{i}=float(10**(-pKa_identified_ions[compound][s]))')
                    #This line below generates symbolic variables representing the concentrations of 
                    #different forms of a compound, where each variable corresponds to a state of dissociation "s" of compound "i" .
                    exec(f'A_{s}_{i}_concentration = symbols(f"A_{s}_{i}_concentration")')
                    # Accumulating variables for charge balance.
                    exec(f'accumulated_variables_for_charge_balance.append((highest_existing_charge_of_identified_ion[compound]-s)*A_{s}_{i}_concentration)')
                    #Accumulating variables for mass balance 
                    exec(f'accumulated_variables_for_mass_balance.append(A_{s}_{i}_concentration)')
                    #Accumulating variables for final proton concentration
                    if s<Degree_of_deprotonation[compound]:
                        exec(f'accumulated_variables_for_final_proton_concentration.append(A_{s}_{i}_concentration)')
                    if s>Degree_of_deprotonation[compound]:
                        exec(f'accumulated_variables_for_final_proton_concentration.append(-1*A_{s}_{i}_concentration)')
                for s in range(number_of_states_of_dissociation):
                    if s!=number_of_states_of_dissociation-1:
                        exec(f'equation_Ka_{s}_{i} = Eq((H_plus_concentration*A_{s+1}_{i}_concentration)/A_{s}_{i}_concentration, Ka_{s}_value_of_compound_{i})')
                        exec(f'equations.append(equation_Ka_{s}_{i})')
                mass_equation_i=Eq(sum(accumulated_variables_for_mass_balance),compound_concentration)
                equations.append(mass_equation_i)
                i+=1
        #Setting up the charge balance equation
        Charge_Balance_equation= Eq(sum(accumulated_variables_for_charge_balance), 0)
        #Adding charge balance equation to the list of equations
        equations.append(Charge_Balance_equation)
        #Setting up the mass balance equation for protons 
        proton_concentration_mass_balance=Eq(sum(accumulated_variables_for_final_proton_concentration),H_plus_zero)
        #Adding mass balance equation for protons to the list of equations
        equations.append(proton_concentration_mass_balance)
        #Setting up the mass balance equation for OH- when there is at least one metal ion in the solution, otherwise, the equation is not needed
        if len(accumulated_variables_for_final_OH_concentration)>1:
            OH_concentration_mass_balance=Eq(sum(accumulated_variables_for_final_OH_concentration),OH_minus_zero)
            equations.append(OH_concentration_mass_balance)
        #Solving the system of equations
        solution = solve(equations, (H_plus_concentration, OH_minus_concentration))
        #If the solution is empty, the pH is set to be a value randomly generated around 6.5 following a normal distribution. This is done to use the function to generate the graph of pH no matter what. 
    except:
        solution = []
    if (len(solution))==0:
        mu = 6.5  # mean
        sigma = 0.2  # standard deviation
        random_pH_value = round(random.gauss(mu, sigma),4)

        return random_pH_value
    pH=-math.log10(solution[0][0])
    return pH


#Dictionary of Ksp_values of metal hydroxides associated to ions 
Ksp_values={
    "Ca": 7.9*10**(-6),
    "Mg": 1.5*10**(-11),
    "K": 10**4, #Complete dissociation of KOH 
    "Mn": 4.6*10**(-14),
    "Zn": 4.5*10**(-17),
    "Cu": 1.6*10**(-19),
    "Fe(3+)": 6.3*10**(-38),}

#Dictionary of pKa_values of compounds
pKa_values={
    "H3PO4": [2.12,7.21,12.32],
    "H3BO3":[9.24,12.4,13.3],
    "HNO3":[-1.4],
    "H2SO4": [-3,1.9],
    "HCl" : [-6.3],
    "NH4" : [9.3],}

#Test with simple case, where expected value is known 
Test_concentration_of_solution = {
    "Ca": float(0),
    "Mg": float(0),
    "K": float(0),
    "PO4": float(0),
    "BO3":float(0),
    "Mn": float(0),
    "Zn": float(0),
    "Cu": float(0),
    "Fe(3+)": float(0),
    "NO3": float(0),
    "H2SO4": float(0.02),
    "Cl" : float(0),
    "NH4" : float(0)}

#expected value of the test
expected_pH=1.58
#Empty dictionary of values of the ionisation constant, the key being the temperature in degrees celcius, and the value being the constant
Value_of_Ionisation_Constant={}
Value_of_Ionisation_Constant[25]=10**(-14)

#Dictionary of charge of metal ion in solution 
compound_charge = {
    "Ca": 2,      # Calcium ion
    "Mg": 2,      # Magnesium ion
    "K": 1,       # Potassium ion
    "Mn": 2,      # Manganese ion
    "Zn": 2,      # Zinc ion
    "Cu": 2,      # Copper(II) ion
    "Fe(3+)": 3   # Iron(III) ion
}

#Dcionary of the highest known charge of the compound when it is undissociated
highest_existing_charge_of_compounds={
    "H3PO4": 0,
    "H3BO3": 0,
    "HNO3": 0,
    "H2SO4": 0,
    "HCl" : 0,
    "NH4" : 1}

def find_acid(ions_concentration:dict, acids_pKa:dict,highest_existing_charge_of_compounds:dict)->tuple:
    """
    Identify acids or their deprotonated forms, and return two dictionaries. One where the keys are the identified ions/acids 
    and attribute to them the pKas of the acid they're associated, the second with the same keys but values being the number of 
    protons the acid would have to lost to reach the form of the ion. 

    Parameters:
    ions_concentration (dict): A dictionary with ion names as keys and their concentrations as values.
    acids_pKa (dict): A dictionary with acid names as keys and their pKa values as lists.

    Returns:
    tuple with three dictionaries 
    1.Dictionary with identified ions being the key and their corresponding values being the number of protons it lost 
    compared to its non-deprotonated form. 
    2.Dictionary where ions are the key and the values are the pkas of the acid
    3.Dictionary where ions are the key and the values are the highest charge of the compound when it is undissociated
    """
    #Create  empty dictionary of number of protons lost by the identified ion/acid
    Degree_of_deprotonation={}
    #Creaty empty dictionary of ions with the corresponding pKa values of the acid they are associated to
    pKa_identified_ions={}
    #Create empty dictionary of the highest known charge of the compound when it is undissociated
    highest_existing_charge_of_identified_ion={}
    #Creates list of acids from pKa dictionary 
    acids = list(acids_pKa.keys())
    for ion, concentration in ions_concentration.items():
        #Value of constant set, which would be zero at end of loop if the ion wasn't identified and one if it was
        #This is used in order to print a message if the ion was not identified if one wants to debug
        i=0
        #Check if the solute is an acid that is not deprotonated
        for acid in acids:
            if ion==acid:
                #print(f"The solute:{ion}, was identified as being an acid in the list of pKa values")
                i=1
                Degree_of_deprotonation[ion]=0
                pKa_identified_ions[ion]=acids_pKa[acid]
                highest_existing_charge_of_identified_ion[ion]=highest_existing_charge_of_compounds[acid]
                continue
            else:
                # Check if the ion/solute is a deprotonated form of any acid
                    if len(ion)!=len(acid):
                        if len(ion)==len(acid)-1:
                            if acid[0] + acid[2:]==ion:
                                #print (f"Ion:{ion} was identified as being the deprotonated form of:", acid)
                                pKa_identified_ions[ion]=acids_pKa[acid]
                                Degree_of_deprotonation[ion]=int(acid[1])-1
                                highest_existing_charge_of_identified_ion[ion]=highest_existing_charge_of_compounds[acid]
                                i=1
                            else:
                                continue
                        elif len(ion)==len(acid)-2:
                            if ion==acid[2:]:
                                #print (f"Ion:{ion} was identified as being the deprotonated form of:", acid)
                                pKa_identified_ions[ion]=acids_pKa[acid]
                                Degree_of_deprotonation[ion]=int(acid[1])
                                highest_existing_charge_of_identified_ion[ion]=highest_existing_charge_of_compounds[acid]
                                i=1
                            else:
                                continue
                    else:
                        if acid[0]+acid[2:]==ion[0]+ion[2:]:
                            i=1
                            #print(f"Ion:{ion} was identified as being the deprotonated form of:", acid)
                            Degree_of_deprotonation[ion]=int(acid[1])-int(ion[1])
                            pKa_identified_ions[ion]=acids_pKa[acid]
                            highest_existing_charge_of_identified_ion[ion]=highest_existing_charge_of_compounds[acid]


        if i!=1: 
            #print("The ion or acid:", ion ,"could not be identified ")
            continue
    return Degree_of_deprotonation, pKa_identified_ions, highest_existing_charge_of_identified_ion

# Test values for the function
Test_concentration_of_solution = {
    "Ca": float(0),
    "Mg": float(0),
    "K": float(0),
    "PO4": float(0),
    "BO3":float(0),
    "Mn": float(0),
    "Zn": float(0),
    "Cu": float(0),
    "Fe(3+)": float(0),
    "NO3": float(0),
    "HSO4": float(0.02),
    "Cl" : float(0),
    "NH4" : float(0)}

#Test
#find_acid(Test_concentration_of_solution, pKa_values)


#Testing the function
#pH_approximation(Test_concentration_of_solution,25)



"""This file contains tests for the functions in the Basic_functions.py file."""

# Importing functions from Basic_functions.py file
import sys
import os
from pytest import approx
sys.path.append(os.path.dirname(os.path.realpath(__file__)) + "/../src")
from hydroponics import *

#--------------------------- Test get_molar_mass() ------------------------------
def test_molar_mass ():
    mol = ["Ca(NO3)2", "Ca(2+)(NO3)2(-)", "Ca++(NO3)2", "Ca(NO3)NO3"]
    for i in range(len(mol)):
        assert round(get_molar_mass(mol[i]),3) == 164.086, "Test molar mass failed: did not understand {mol[i]}"
           


#--------------------------- Test get_Ksp() ------------------------------
def test_get_Ksp():
    list = ['KNO3', 'KH2PO4', 'MnCl2', 'NaCl', 'KCl', 'KOH', 'K2SO4', 'K2CO3', 'KHCO3', 'CaCl2', 'MgCl2', 'FeCl3', 'AlCl3']
    Ksp_list = [
        9.76904463917891,
        2.758051364759525,
        810.1350827255483,
        37.716373801958376,
        21.04634685946477,
        398.49994656470716,
        1.0339299769044514,
        2072.3420181206798,
        11.330982570493006,
        1210.091270423005,
        754.4989632174183,
        0,
        3759.188606917454
    ]
    for i in range(len(list)):
        assert get_Ksp(list[i]) == approx(Ksp_list[i], rel=1e-3), f"Test_get_Ksp failed: Error for {list[i]}"
        
        

#----------------------------- Test get_Q() ----------------------------
def test_Get_Q():
    salt = "Ca(NO3)2"
    ions = {"Ca(2+)": 0.5, "NO3(-)": 0.5, "Cl-": 0.5}
    assert get_Q_solubility(salt, ions) == 0.125, "Test get_Q failed"
    
    
#----------------------------- Test salt2ions() ----------------------------
def test_mol_unit():
    solution_dict = {"NaCl": 0.1, "KBr": 0.05} # 0.1 mol/L NaCl, 0.05 mol/L KBr
    assert salt2ions(solution_dict, volume=2, unit='mol') == {'Na+': 0.2, 'Cl-': 0.2, 'K+': 0.1, 'Br-': 0.1}, "Salt2Ions: Test_mol_unit failed"
    
def test_g_unit():
    solution_dict = {"NaCl": 0.1, "KBr": 0.05} # 0.1 g/L NaCl, 0.05 g/L KBr
    assert salt2ions(solution_dict, volume=2, unit='g') == {'Na+': 4.597953856, 'Cl-': 7.090000000000001, 'K+': 3.9098300000000004, 'Br-': 7.9904}, "Salt2Ions: Test_g_unit failed"

def test_invalid_volume():
    solution_dict = {"NaCl": 0.1, "KBr": 0.05}
    try:
        salt2ions(solution_dict, volume=0)
    except ValueError as e:
        assert "{}".format(e) == "Volume must be positive.", "Salt2Ions: invalid volume failed"
        

#----------------------------- Test make_solution() ----------------------------
def test_make_solution():
    ions_in_solution = {"K+": 0.5, "Cl-": 0.6, "Br-": 0.1, "SO4(2-)": 0.5 }
    forbidden_ions = ['NO3(-)']
    volume = 10
    expected_result = {'KH2PO4': 15.6997161195170, 'MnCl2': 10.6491996897038, 'KBr': 1.48931592911494, 'MgSO4': 6.26514741400849}
    assert make_solution(ions_in_solution, forbidden_ions, volume) == approx(expected_result, 1e-3), "Test make_solution failed"
#----------------------------- Test find_acid() ----------------------------

def test_find_acid():
    pKa_values={
    "H3PO4": [2.12,7.21,12.32],
    "H3BO3":[9.24,12.4,13.3],
    "HNO3":[-1.4],
    "H2SO4": [-3,1.9],
    "HCl" : [-6.3],
    "NH4" : [9.3],}

    Test_concentration_of_solution = {
        "Ca": float(0),
        "Mg": float(0),
        "K": float(0),
        "PO4": float(0),
        "BO3":float(0),
        "Mn": float(0),
        "Zn": float(0),
        "Cu": float(0),
        "Fe(3+)": float(0),
        "NO3": float(0),
        "HSO4": float(0.02),
        "Cl" : float(0),
        "NH4" : float(0)}
    
    highest_existing_charge_of_compounds = {'H3PO4': 0, 'H3BO3': 0, 'H2SO4': 0, 'NH4': 1}
    
    expected_dict=({'PO4': 3, 'BO3': 3, 'HSO4': 1, 'NH4': 0},
    {'PO4': [2.12, 7.21, 12.32],
    'BO3': [9.24, 12.4, 13.3],
    'HSO4': [-3, 1.9],
    'NH4': [9.3]},{'PO4': 0, 'BO3': 0, 'HSO4': 0, 'NH4': 1})
    assert find_acid(Test_concentration_of_solution,pKa_values,highest_existing_charge_of_compounds) == expected_dict

#----------------------------- Test pH_approximation() ----------------------------
def test_pH_approximation():
    pKa_values={
    "H3PO4": [2.12,7.21,12.32],
    "H3BO3":[9.24,12.4,13.3],
    "HNO3":[-1.4],
    "H2SO4": [-3,1.9],
    "HCl" : [-6.3],
    "NH4" : [9.3],}
    
    Test_concentration_of_solution = {
        "Ca": float(0),
        "Mg": float(0),
        "K": float(0),
        "PO4": float(0),
        "BO3":float(0),
        "Mn": float(0),
        "Zn": float(0),
        "Cu": float(0),
        "Fe(3+)": float(0),
        "NO3": float(0),
        "HSO4": float(0.02),
        "Cl" : float(0),
        "NH4" : float(0)}
    
    result = pH_approximation(Test_concentration_of_solution,25)
    lower_bound = 6
    upper_bound = 7.5
    assert lower_bound <= result <= upper_bound, f"Test failed: {result} is not within the range {lower_bound} to {upper_bound}"

#----------------------------- Test Refill_of_container() ----------------------------
def test_Refill_of_container():
    concentration_of_ions = {"Na+": 0.1, "Cl-": 0.1, "K+": 0.05, "NO3(-)": 0.05}
    optimal_ion_concentrations = {"Na+": 0.2, "Cl-": 0.2, "K+": 0.1, "NO3(-)": 0.1}
    assert Refill_of_container(concentration_of_ions, optimal_ion_concentrations, 5) == {'Na+': 0.1, 'Cl-': 0.1, 'K+': 0.05, 'NO3(-)': 0.05}, "Refill_of_container: Test failed"

#----------------------------- Test generation_of_pH_list() ----------------------------
def test_generation_of_pH_list():
      Test_concentrations_list_of_solution =[{
          "Ca": float(0),
          "Mg": float(0),
          "K": float(0),
          "PO4": float(0),
          "BO3":float(0),
          "Mn": float(0),
          "Zn": float(0),
          "Cu": float(0),
          "Fe(3+)": float(0),
          "NO3": float(0),
          "HSO4": float(0.02),
          "Cl" : float(0),
          "NH4" : float(0)},{
          "Ca": float(0),
          "Mg": float(0),
          "K": float(0),
          "PO4": float(0),
          "BO3":float(0),
          "Mn": float(0),
          "Zn": float(0),
          "Cu": float(0),
          "Fe(3+)": float(0),
          "NO3": float(0),
          "HSO4": float(0.02),
          "Cl" : float(0),
          "NH4" : float(0)}]
      temperature = 25
      plant = "Eggplant"
      list_of_pH,exceeded_days_list=generation_of_pH_list(Test_concentrations_list_of_solution,temperature,plant)
      for i in range (len(list_of_pH)):
          assert 6.0<=list_of_pH[i]<= 8.0, "generation_of_pH_list: Test failed"
      assert type(exceeded_days_list) == list, "generation_of_pH_list: Test failed"


"""
This file contains the most basic functions used in the project.
In particular, the file contains functions to:
- Import user data from an Excel file
- Calculate the molar mass of a salt
- Get the solubility product constant (Ksp) and solubility product (Q) of a salt
- Convert salt concentrations to ion concentrations
- Make a solution given the ion concentration wanted
"""

#Imports
import re
import os
import csv
import pandas as pd
from sympy import symbols, Eq, solve, solveset, S, Interval 


# ----- Import User Data from Excel file -------------

def load_from_data(file_name):
    # Get the directory of the current file
    current_dir = os.path.dirname(os.path.realpath(__file__))
    # Construct the absolute path to the CSV file
    final_path = current_dir+'\data\\'+ file_name
    return final_path

def import_solution_data(solution_name: str) -> dict:
    """
    Import user data from the Excel file "UserData.xlsx" and returns it as a dictionary.

    Args:
        solution_name (str): Name of the column in the Excel file.

    Returns:
        dict: Dictionary containing the concentration of the ions [g/L] in the solution.
    """
    # Read the Excel file
    file_name = load_from_data("UserData.xlsx")
    df = pd.read_excel(file_name, sheet_name="My solution", skiprows=1)
    # Create a dictionary using zip and dictionary comprehension
    try:
        ion_solution_dict = dict(zip(df['Ions'], df[solution_name]))
    except KeyError:
        raise ValueError(f"Solution {solution_name} not found in the Excel file. Please check the name of the column.")
    return ion_solution_dict

def predefined_solutions(plant:str, concentration:str="g" ) -> dict:
    """
    Import user data from the Excel file "UserData.xlsx" and returns it as a dictionary.

    Args:
        plant (str): Name of the plant. Can either be 'Eggplant', 'Tomato' or 'Cucumber'.
        concentration (str): Unit of concentration. Can either be 'g' or 'mol'.

    Returns:
        dict: Dictionary containing the concentration of the ions [g or mol/L] in the solution.
    """
    if concentration not in ['g','mol']:
        raise ValueError("Concentration must be 'g' or 'mol'.")
    available_plants = ['Eggplant', 'Tomato', 'Cucumber']
    if plant not in available_plants:
        raise ValueError(f"Plant {plant} not found. Available plants are {available_plants}.")
    else:
        # Read the Excel file
        file_name = load_from_data("UserData.xlsx")
        df = pd.read_excel(file_name, sheet_name="Optimal Solution", skiprows=0)
        # Create a dictionary using zip and dictionary comprehension
        ion_solution_dict = dict(zip(df['Formula'], df[str('c ['+concentration+'/L] '+plant)]))
        return ion_solution_dict

def import_plant_data(plant_name:str) -> dict:
    """
    Import user data from sheet "My Plant" from the Excel file "UserData.xlsx" and returns it as a dictionary.

    Args:
        plant (str): name of the column where the data is located
        
    Returns:
        dict: dictionary containing the needs of the plant for each ion [g] for a full growth cycle
    """
    # Read the Excel file
    file_name = load_from_data("UserData.xlsx")
    df = pd.read_excel(file_name, sheet_name="My plant", skiprows=1)
    # Create a dictionary using zip and dictionary comprehension
    plant_dict = dict(zip(df['Ions'], df[plant_name]))
    return plant_dict



# ----- Molar mass calculator ------------------------

def get_atom_mass(atom_name: str) -> float:
    """
    Gets the atomic mass of the atom from a CVS file

    Args:
        atom_name (str): Name of the atom, i.e. 'H', 'O', 'C', etc.

    Returns:
        float: atomic mass of the atom in g/mol
    """
    # Define the name of the CSV file
    data_file = load_from_data("molar_mass.csv") #Check path to CSV file
    with open(data_file, mode='r') as file:
        reader = csv.DictReader(file)
        for row in reader:
            if row['Atom'] == atom_name:
                return float(row['Molar mass [g/mol]'])

def get_molar_mass(salt: str) -> float:
    """
    Calculate the molar mass of a salt in g/mol.
    
    Args:
        salt (str): The chemical formula of the salt.
            Handles chemical formulas of the following types:
            - "Ca(NO3)2"
            - "Ca(2+)(NO3-)2"
            - "Ca2(NO3)2"
            - "Ca(NO3)2(2+)"
            Please do not use spaces in the formula and avoid putting brackets inside brackets.

    Returns:
        float: The molar mass of the salt in g/mol.
    """
    #Check for special characters
    def contains_special_characters(input_string:str):
        # Define a regular expression pattern to match special characters
        pattern = re.compile(r'[!@#$%^&*_=\[\]{};\'\\:"|,.<>\/?]')
        # Use search method to find if any special characters exist in the input string
        if pattern.search(input_string):
            return True
        else:
            return False

    def molar(molecule:str):
        #Check for special characters
        if contains_special_characters(molecule):
            raise ValueError("Invalid input. The formula of the molcule can only contain letters,numbers, parentheses and '+' or '-' signs.")
        
        current_count = ""
        current_element = ""
        i=0
        molar_mass = 0
        #Test for charges and numbers
        while molecule[-1] == "+" or molecule[-1] == "-":
            molecule = molecule[:-1]
            if len(molecule) == 0:
                break
        if molecule.isnumeric():
            return 0   
    
        while i < len(molecule):
            if i<len(molecule)-3:
                if molecule[i].isupper() and molecule[i+1].islower() and molecule[i+2].isnumeric() and molecule[i+3].isnumeric():
                    current_element = molecule[i]+molecule[i+1]
                    current_count = molecule[i+2]+molecule[i+3]
                    i+=4
                    molar_mass += get_atom_mass(current_element)*int(current_count)
                    continue
                
            if i<len(molecule)-2:
                if molecule[i].isupper() and molecule[i+1].islower() and molecule[i+2].isnumeric():
                    current_element = molecule[i]+molecule[i+1]
                    current_count = molecule[i+2]
                    i+=3
                    molar_mass += get_atom_mass(current_element)*int(current_count)
                    continue
                    
                if molecule[i].isupper() and molecule[i+1].islower() and molecule[i+2].isalpha():
                    current_element = molecule[i]+molecule[i+1]
                    current_count = 1
                    i+=2
                    molar_mass += get_atom_mass(current_element)*int(current_count)
                    continue
                
                if molecule[i].isupper() and molecule[i+1].isnumeric() and molecule[i+2].isnumeric():
                    current_element = molecule[i]
                    current_count = molecule[i+1]+molecule[i+2]
                    i+=2
                    molar_mass += get_atom_mass(current_element)*int(current_count)
                    continue
            if i<len(molecule)-1:
                if molecule[i].isupper() and molecule[i+1].isnumeric():
                    current_element = molecule[i]
                    current_count = molecule[i+1]
                    i+=2
                    molar_mass += get_atom_mass(current_element)*int(current_count)
                    continue
                elif molecule[i].isupper() and molecule[i+1].isupper():
                    current_element = molecule[i]
                    current_count = 1
                    i+=1
                    molar_mass += get_atom_mass(current_element)*int(current_count)
                    continue
            
            if i == len(molecule)-2:
                if molecule[i].isupper() and molecule[i+1].islower():
                    current_element = molecule[i]+molecule[i+1]
                    current_count = 1
                    i+=2
                    molar_mass += get_atom_mass(current_element)*int(current_count)
                    continue
                if molecule[i].isupper() and molecule[i+1].isupper():
                    current_element = molecule[i]
                    current_count = 1
                    i+=1
                    molar_mass += get_atom_mass(current_element)*int(current_count)
                    continue
                if molecule[i].isupper() and molecule[i+1].isnumeric():
                    current_element = molecule[i]
                    current_count = molecule[i+1]
                    i+=2
                    molar_mass += get_atom_mass(current_element)*int(current_count)
                    continue  
            if i == len(molecule)-1:
                current_element = molecule[i]
                current_count = 1
                i+=1
                molar_mass += get_atom_mass(current_element)*int(current_count)
                continue
            if molar_mass >=100000:
                break

        return molar_mass
    
    #Take care of brakets
    M = 0    
    while "(" in salt:
        left = salt.index("(")
        right = salt.index(")")
        substing = salt[left+1:right]

        if right != len(salt)-1:
            if salt[right+1].isnumeric():
                M += int(salt[right+1])*molar(substing)
                salt = salt[:left] + salt[right+2:]
            else:
                M += molar(substing)
                salt = salt[:left] + salt[right+1:]
        else:
            M += molar(substing)
            salt = salt[:left] + salt[right+1:]
            
    total_molar_mass = M + molar(salt)
    return total_molar_mass



#------ Basic Solubility functions -----------------------------
#Dictionaries used to handle salts with multiple ions, extend dictionaries if needed
salt2nbIons = {
    "Ca(NO3)2": [1,2],
    "MgSO4": [1,1],
    "KNO3": [1,1],
    "KH2PO4": [1,1],
    "H3BO3": [1],
    "MnCl2": [1,2],
    "ZnSO4": [1,1],
    "CuSO4": [1,1],
    "(NH4)6Mo7O24": [6,1],
    "C10H12FeN2NaO8": [1,1,1],
    "Fe(III)EDTANa": [1,1,1], #
    "(NH4)H2PO4" : [1,1],
    "NaCl": [1,1],
    "NaOH": [1,1],
    "Na2CO3": [2,1],
    "Na2SO4": [2,1],
    "NaHCO3": [1,1],
    "Na2S": [2,1],
    "KCl": [1,1],
    "KOH": [1,1],
    "K2SO4": [2,1],
    "K2CO3": [2,1],
    "KHCO3": [1,1],
    "CaCl2": [1,2],
    "Ca(OH)2": [1,2],
    "CaCO3": [1,1],
    "MgCl2": [1,2],
    "Mg(OH)2": [1,2],
    "MgCO3": [1,1],
    "FeSO4": [1,1],
    "FeCl3": [1,3],
    "AlCl3": [1,3],
    "KBr" : [1,1],
    "Mg(NO3)2": [1,2],
    "Zn(NO3)2": [1,2],
    "Cu(NO3)2": [1,2]

    
}
dict_salts_trad = {
    "Ca(NO3)2": {"Ca(2+)": 1, "NO3(-)": 2},
    "MgSO4": {"Mg(2+)": 1, "SO4(2-)": 1},
    "KNO3": {"K+": 1, "NO3(-)": 1},
    "KH2PO4": {"K+": 1, "H2PO4(-)": 1},
    "H3BO3": {"B": 1},
    "MnCl2": {"Mn(2+)": 1, "Cl-": 2},
    "ZnSO4": {"Zn(2+)": 1, "SO4(2-)": 1},
    "CuSO4": {"Cu(2+)": 1, "SO4(2-)": 1},
    "(NH4)6Mo7O24": {"NH(4+)" : 6, "Mo": 1},
    "C10H12FeN2NaO8": {"Fe(3+)": 1, "Na+": 1, "EDTA": 1},
    "Fe(III)EDTANa": {"Fe(3+)": 1, "Na+": 1, "EDTA": 1},
    "(NH4)H2PO4" : {"NH4+": 1, "H2PO4(-)": 1},
    "NaCl": {"Na+": 1, "Cl-": 1},
    "NaOH": {"Na+": 1, "OH-": 1},
    "Na2CO3": {"Na+": 2, "CO3": 1},
    "Na2SO4": {"Na+": 2, "SO4(2-)": 1},
    "NaHCO3": {"Na+": 1, "HCO3(-)": 1},
    "Na2S": {"Na+": 2, "S(2-)": 1},
    "KCl": {"K+": 1, "Cl-": 1},
    "KOH": {"K+": 1, "OH-": 1},
    "K2SO4": {"K+": 2, "SO4(2-)": 1},
    "K2CO3": {"K+": 2, "CO3(2-)": 1},
    "KHCO3": {"K+": 1, "HCO3(-)": 1},
    "CaCl2": {"Ca(2+)": 1, "Cl-": 2},
    "Ca(OH)2": {"Ca(2+)": 1, "OH-": 2},
    "CaCO3": {"Ca(2+)": 1, "CO3(2-)": 1},
    "MgCl2": {"Mg(2+)": 1, "Cl-": 2},
    "Mg(OH)2": {"Mg(2+)": 1, "OH-": 2},
    "MgCO3": {"Mg(2+)": 1, "CO3(2-)": 1},
    "FeSO4": {"Fe(2+)": 1, "SO4(2-)": 1},
    "FeCl3": {"Fe(3+)": 1, "Cl-": 3},
    "AlCl3": {"Al(3+)": 1, "Cl-": 3},
    "KBr" : {"K+": 1, "Br-": 1},
    "Mg(NO3)2": {"Mg(2+)": 1, "NO3(-)": 2},
    "Zn(NO3)2": {"Zn(2+)": 1, "NO3(-)": 2},
    "Cu(NO3)2": {"Cu(2+)": 1, "NO3(-)": 2}

    
}


# Get the solubility constant Ksp of a salt
def get_Ksp(salt_name: str) -> float:
    """
    Access the solubility product constant (Ksp) of a salt based on its name. 
        Uses a salt2nbIons dictionary to find the number of ions in a salt.

    Args:
        salt_name (str): The formula of the salt. Use standard notation for the formula,
                        e.g. "Ca(NO3)2", "H2O", etc.

    Returns:
        float: The solubility product constant (Ksp) of the salt.
    """
    solubility_file = load_from_data("Solubility_data.csv")
    df_solubility = pd.read_csv(solubility_file)
    
    row = df_solubility.loc[df_solubility["Formula"] == salt_name].values.tolist() # [[name, formula, value]]
    Ksp = 0
    if row:
        solubility = float(row[0][2]) *10 #go from g/100mL to g/L
        mol_sol = solubility/get_molar_mass(salt_name) #mol/L
        
        #Ksp = n^n * m^m * mol_sol^(n+m) for salt of type nXmY
        ions = salt2nbIons[salt_name]
        if len(ions) <=1:
            Ksp = mol_sol
        else: 
            n = ions[0]
            m = ions[1]
            Ksp = (n**n)*(m**m)*(mol_sol**(n+m))
    return Ksp 

# Get the solubility product Q of a salt
def get_Q_solubility(salt_name: str, ions_in_solution: dict) -> float:
    '''
    Returns the solubility product of a salt
    
    Args:
        salt_name (str): name of the salt
        ions_in_solution (dict): dictionary with keys as ions and values as the concentration in mol/L
    
    Returns:
        float: solubility product Q
    '''
    Q = 1
    i=0
    for ion in dict_salts_trad[salt_name]:
        if ion in ions_in_solution:
            Q *= ions_in_solution[ion]**salt2nbIons[salt_name][i]
            i+=1
            
    return Q




#-----------Basic ion-salt functions----------------------------

# Convert salt concentrations to ion concentrations
def salt2ions(solution: dict, volume: float = 1, unit: str = "g") -> dict:
    """
    Convert salts to ions based on their concentrations and volume. Concentrations are in g/L by default 
        but can also be given in mol/L by specifying 'unit' as 'mol'.
        A 'dict_salts_trad' dictionary is used to convert salt names to ions.
        Make sure the salts are in the 'dict_salts_trad' dictionary.

    Args:
        solution (dict): A dictionary containing salt names as keys and concentrations as values.
        volume (float, optional): Volume of the solution. Defaults to 1.
        unit (str, optional): Unit of concentration. Defaults to "g". Options are "g" or "mol".

    Returns:
        dict: A dictionary containing ions and their concentrations [g/L] in the solution (volume =1), their quantity [g] if volume > 1

    Raises:
        ValueError: If the volume is non-positive.
    """
    if volume <= 0:
        raise ValueError("Volume must be positive.")
    if unit == "mol":
        ions = {}
        for salt in solution:
            if salt in dict_salts_trad:
                for ion in dict_salts_trad[salt].keys():
                    if ion in ions:
                        ions[ion] += dict_salts_trad[salt][ion] * solution[salt]
                    else:
                        ions[ion] = dict_salts_trad[salt][ion] * solution[salt]
            else:
                raise ValueError(f"Salt {salt} not found in the 'dict_salts_trad' dictionary.")
            
        ions_in_solution = {n: volume * ions[n] for n in ions.keys()}    
        return ions_in_solution
    
    elif unit == "g":
        salts = {salt: mass/get_molar_mass(salt) for salt, mass in solution.items()}
        ions = {}
        for salt in salts:
            if salt in dict_salts_trad:
                for ion in dict_salts_trad[salt].keys():
                    if ion in ions:
                        ions[ion] += dict_salts_trad[salt][ion] * solution[salt]
                    else:
                        ions[ion] = dict_salts_trad[salt][ion] * solution[salt]
            else:
                raise ValueError(f"Salt {salt} not found in the 'dict_salts_trad' dictionary.")
        ions_in_solution = {ion: float(get_molar_mass(ion)) * float(volume) * ions[ion] for ion in ions.keys()}
        return ions_in_solution
    else:
        raise ValueError("'Unit' must be 'g' or 'mol'.")

#Make a solution given the ion concentration wanted
def make_solution (ions_in_solution :dict, forbidden_ions:list, volume:float, )->dict:
    '''
    Returns how much of each salt [g] to add to a solution of certain volume [L]
    to obtain a certain concentration of ions [g/L]
    
    Args:
        ions_in_solution (dict): dictionary with keys as ions and values as the concentration in g/L.
        forbidden_ions (list): list of ions that cannot be in the solution.
        volume (float): volume of the final solution in L.
    
    Returns:
        dict: dictionary with keys as salt names and values as the amount of salt needed in g.
    '''
    
    if set(ions_in_solution.keys()).intersection(forbidden_ions) != set() :
        problems = set(ions_in_solution.keys()).intersection(forbidden_ions)
        raise ValueError(f"Forbidden ions are required in the solution: {problems}")
    
    molar_ions = {ion: ions_in_solution[ion]/get_molar_mass(ion) for ion in ions_in_solution}
    # Define variables
    possible_salts = []
    for salt in dict_salts_trad.keys():
        for ion in dict_salts_trad[salt]:
            if (ion in molar_ions) and (salt not in possible_salts) and (len(set(dict_salts_trad[salt].keys()).intersection(forbidden_ions)) == 0):
                possible_salts.append(salt)
    
    #Check for soluble salts
    for salt in possible_salts:
        if get_Ksp(salt) <= get_Q_solubility(salt, molar_ions):
            possible_salts.remove(salt)
    
    #NB variables == NB equations
    final_salts = []
    for ion in molar_ions.keys():
        for salt in possible_salts:
            if ion in dict_salts_trad[salt] and salt not in final_salts:
                final_salts.append(salt)
                break
                
    variables = symbols(final_salts)
        
    #Set up equations
    equations = []
    for ion in molar_ions.keys():
        lhs = 0
        rhs = molar_ions[ion] * volume
        for salt in final_salts:
            if ion in dict_salts_trad[salt]:
                lhs += dict_salts_trad[salt][ion] * symbols(salt)
        equation = Eq(lhs, rhs)
        equations.append(equation)
    
    #Solve equations
    try :
        positive_solutions = solveset(equations, variables, domain=Interval(0, 1000000))
    except:
        positive_solutions = set()
    if positive_solutions == set():
        solution = solve(equations, variables)
        if solution == []:
            raise ValueError("No solution found. Please check the ions in the solution and the forbidden ions.")
        else:
            mass_salts = {salt: solution[symbols(salt)]*get_molar_mass(salt) for salt in final_salts}
            return mass_salts  
    else:
        mass_salts = {salt: positive_solutions[symbols(salt)]*get_molar_mass(salt) for salt in final_salts}
        return mass_salts
        
        



"""
This file contains the more advanced functions on solutions and solubility used in the project.
In particular, the file contains functions to:
- Check the solubility of a solution
- Analyse the nutriments in the solution
- Update the solution after 1 day of growth
- Plot the evolution of the solution
"""

#Imports
import pandas as pd
import math
import matplotlib.pyplot as plt
import os
import numpy as np
from .Basic_functions import get_molar_mass, get_Ksp, get_Q_solubility, salt2ions, dict_salts_trad

# ----------------- Analyse the solution -----------------

#Check the solubility of the solution
def check_solubility(salts_dict: dict, input_type: str = "salt", output_type: str = "bool") -> bool:
    """
    Check if the salts mixture is soluble.

    Args:
        salts_dict (dict): Dictionary with keys as salt names and values as concentrations [g/L].
        input_type (str): Type of input data, either "salt" or "ion".
        output_type (str): Type of output data, either "bool", "analysis", or "update".

    Returns:
        bool or str: True if salts are soluble (if output_type is "bool").
                     Analysis of precipitated salts or indication of solubility (if output_type is "analysis").
    """
    # Validate input and output types
    valid_input_types = {"salt", "ion"}
    valid_output_types = {"bool", "analysis", "update"}
    if input_type not in valid_input_types:
        raise ValueError("Invalid input type. Please choose either 'salt' or 'ion'.")
    if output_type not in valid_output_types:
        raise ValueError("Invalid output type. Please choose either 'bool', 'analysis', or 'update'.")
    
    #Handle type of input and convert to mol/L
    if input_type == "salt":
        salts_dict_molar= {salt: concentration / get_molar_mass(salt) for salt, concentration in salts_dict.items()}
        ions = salt2ions(salts_dict_molar, volume = 1)
    if input_type == "ion":
        ions = {ion: concentration / get_molar_mass(ion) for ion, concentration in salts_dict.items()}
        
    #Initialize variables
    soluble = True
    precipitate = []
    salts = dict_salts_trad.keys()
    
    #Compare Q and Ksp for each salt is the Ksp value is provided
    for salt in salts:
        try:
            Ksp = get_Ksp(salt)
            Q = get_Q_solubility(salt, ions)
            if Q > Ksp and set(dict_salts_trad[salt].keys()).issubset(set(ions.keys())):
                soluble = False
                precipitate.append(salt)
        except:
            pass
            
    
    #Handle type of output
    if output_type == "bool":
        return soluble
    elif output_type == "analysis":
        if soluble:
            return "all salts are soluble"
        else:
            text = f"The following salts precipitate: "
            for salt in precipitate:
                text+=str(salt)+"    "
            return text
    elif output_type == "update":
        return precipitate
    
#analyse the nutriments in the solution
def analyse_nutriments(solution: dict, plant: dict, growth_time: float, volume: float, input_type_solution: str= 'salt') -> list:
    """
    Returns the number of days a plant can grow with the given solution.

    Args:
        solution (dict): A dictionary containing salts or ions and their concentrations in the solution.
        input_type (str): Type of input data, either "salt" or "ion".
        plant (dict): A dictionary containing ions and their required quantity for plant growth.
        growth_time (float): Expected growth time of the plant [days].
        volume (float): Volume of the solution [L].
        *Note: the unit of solution and plant must be the same. (either mol/L or g/L)

    Returns:
        list: A list containing three elements:
            - bool: True if the solution contains enough nutriments for the plant, False otherwise.
            - int: Minimum days of growth achievable with the given solution.
            - str: Ion that limits the plant growth.

    Raises:
        ValueError: If the volume or growth time is non-positive.
    """
    #Check validity of Args
    if volume <= 0:
        raise ValueError("Volume must be positive.")
    if growth_time <= 0:
        raise ValueError("Growth time must be positive.")
    if input_type_solution not in {"salt", "ion"}:
        raise ValueError("Invalid input type. Please choose either 'salt' or 'ion'.")
    
    #Initialisation
    if input_type_solution == "salt":
        ions = salt2ions(solution, volume=volume)
    if input_type_solution == "ion":
        ions = {ion: solution[ion]*volume for ion in solution.keys()}
        
    daily_plant_need = {ion: plant[ion]/growth_time for ion in plant.keys()}
    limiting_ion = None
    growth_limit = growth_time
    
    #Determine limiting ion, minimum days of growth and if growth time exceeds the minimum days
    for ion in ions:
        if ion in daily_plant_need:
            days = math.floor(ions[ion] / daily_plant_need[ion])
            if days < growth_limit:
                growth_limit = days
                limiting_ion = ion
    if set(plant.keys()).issubset(set(ions.keys())):
        return [growth_limit == growth_time, growth_limit, limiting_ion]
    else:
        missing_ions = [ion for ion in plant.keys() if ion not in ions]
        return [False, 0, missing_ions]
   
#Check is enaugh nutriments are present in the solution
def check_supply_elements(ions_solution, plant) -> bool:
    for ion in plant.keys():
        if ion not in ions_solution:
            raise ValueError("{ion} is not present in the solution but is required for the plant growth.")
        elif plant[ion] > ions_solution[ion]:
            return False
    return True



# ----------------- Evolution of the solution -----------------

#Update the solution after 1 day of growth
def update_sol(ions_solution: dict, plant: dict, volume: float) -> dict:
    """
    updates the salts in solutions after 1 day of growth and checks the solubility of the new solution
    
    Args:
        ions_solution: dictionary with keys as ions and values as the amount of ions in the solution [g/L]
        plant: dictionary with keys as ions and values as the amount of ions needed for 1 day for the plant [g]
        
    Returns:
        dictionary: updated ions_solution after 1 day of growth and checking the solubility
    """
    volume_plant = {ion: plant[ion]/volume for ion in plant.keys()}
    for ion in volume_plant:
        if ions_solution[ion] <= volume_plant[ion]:  
            ions_solution[ion] = 0
        else:      
            ions_solution[ion] -= volume_plant[ion]
    
    if check_solubility(ions_solution, input_type="ion"):
        return ions_solution  
    else: 
        unsoluble_ions = check_solubility(ions_solution, input_type ="ion", output_type = "update")
        for ion in unsoluble_ions:
            ions_solution[ion] = 0
        return ions_solution
    
# Evolution of the concentration as the plant grows (internal function)
def data4graph(solution: dict, volume: float, plant: dict, growth_time: float) -> list:
    """
    Creates a dictionary with the data needed to plot the graph
    
    Args:
        solution (dict): Initial concentration [g/L] of the ions in the solution.
        volume (float): Volume of the solution [L].
        plant (dict): Dictionary containing required amount of ions for the plant growth.
        growth_time (float): Expected growth time of the plant [days].
    
    Returns:
        list: A list containing two elements:
            - list: A list of days. [0,1,2,3,4,5,6,...]
            - list: A list of dictionaries with the concentration of ions in the solution for each day.
    
    """
    #Initialisation:
    daily_need_plant = {ion: plant[ion]/growth_time for ion in plant.keys()}
    data = [[],[]] #data = [[0,1,2,3,4,5,6,...],[{"Na",0.1,"K",0.2,...}, {"Na": 0.05, "K": 0.1,...},...]]
    i=0
    data[0].append(i)
    data[1].append(solution.copy())
    
    #Loop trough further values
    for i in range(1,growth_time+1):
            if check_supply_elements(solution, daily_need_plant):
                solution = update_sol(solution, daily_need_plant, volume)
                data[0].append(i)
                data[1].append(solution.copy())
            else:
                data[0].append(i)
                data[1].append(solution.copy())
        
    return data
    
#List of colors for the graph
python_colors = ['b', 'g', 'c', 'm', 'y', 'k',
                 'tab:blue', 'tab:orange', 'tab:green', 'tab:red', 'tab:purple', 'tab:brown', 'tab:pink', 'tab:gray', 'tab:olive', 'tab:cyan',
                 'mediumblue', 'darkorange', 'limegreen', 'indianred', 'violet', 'sienna', 'deepskyblue', 'peru', 'gold']

#Plot the evolution of the solution
def plot_graph(solution: dict, input_type:str, plant:dict, growth_time:float, volume:float = 1, ions_of_interest: list = "all"):
    """
    Creates graph of salts in hydroponic solution

    Args:
        salts (dict): Initial concentration [g/L] of the ions in the solution.
        input_type (str): Type of input data, either "salt" or "ion".
        plant (dict): the plant requirements to fully grow [g].
        growth_time (float): Expected growth time of the plant [days].
        volume (float, optional): Volume of the solution [L]. Defaults to 1.
        ions_of_interest (list, optional): list of ions to plot. Defaults to "all". 
            Elements must be part of the keys of solution.
    
    Returns:
        None, saves the graph as a .png file in the download folder.
    """
    
    #Initialise the figure and aesthetics
    plt.figure(figsize=(12, 6), dpi = 500)
    plt.title('Evolution of the Ions in hydroponic solution', fontsize=16, weight='bold')
    plt.xlabel(f'Time [days]', fontsize=14)
    plt.ylabel('Concentration of the ion [g/L]', fontsize=14)
    plt.grid(False)
    plt.gca().set_facecolor('#f9f9f9')
    plt.gca().spines['top'].set_visible(False)
    plt.gca().spines['right'].set_visible(False)
    
    # Setting y-axis to log scale
    #plt.yscale('log')
    
    
    #xlength = analyse_nutriments(solution, plant, growth_time, volume, input_type_solution = 'ion')[1]
    
    #Handle Input type
    if input_type == "salt":
        initial_ion_solution = salt2ions(solution, volume)
    else:
        initial_ion_solution = solution.copy()
        
    #add the growth limit
    analysis = analyse_nutriments(initial_ion_solution, plant, growth_time, volume, input_type_solution = 'ion')
    if analysis[1] != growth_time:
        plt.axvline(x=analysis[1], color='r', linestyle='--', label=f'Growth limit: insufficient {analysis[2]}')
        
    #Plot the data for the ions of interest
    i = 0 #handle the colors    
    data = data4graph(initial_ion_solution, volume, plant, growth_time)
    for ion in initial_ion_solution.keys():
        if ion in ions_of_interest or ions_of_interest == "all":
            x = data[0]
            y = []
            for j in range(len(x)):
                y.append(data[1][j][ion])
            plt.plot(x, y, label = ion, color = python_colors[i],alpha=0.8, linewidth=1.5, linestyle='-', marker='o', markersize=3)
            i+=1
            
    
    plt.legend(loc='best')
    
    #Save Graph to the current directory
    current_dir = os.path.dirname(os.path.realpath(__file__))
    file_name = "graph_"
    for ion in ions_of_interest:
        file_name += str(ion) + "_"
    file_path = os.path.join(current_dir, file_name)
    plt.savefig(file_path + ".png", format='png')  # Save as a PNG file for matplotlib
    
 


import matplotlib.pyplot as plt  # Importing matplotlib for plotting
import pandas as pd
from hydroponics.PH_Approximation_0 import pH_approximation
import os

#Dictionary with optimal range for plants to grow in 
pH_limit_of_plants = {
    "Eggplant": [6.2, 6.8],
    "Cucumber": [6.0, 7.0],
    "Bell pepper": [5.5, 6.5]
}

def generation_of_pH_list(concentrations_list:list, temperature:float, plant:str)->(tuple):
    """
    Generate pH values and identifies days where pH exceeded limits.

    Parameters:
        concentrations_list (list): list of dictionaries of concentrations throughout days, with the index of the 
        dictionary in the list being the day where the concentrations were derived.
        temperature (float): Temperature of the solution in degrees celcius.
        plant (str): Name of the plant.

    Returns:
        list: pH values for each concentration dictionary.
        list: Indices of days where pH exceeded limits.
    """
    pH_values = []  # List to store pH values
    Days_where_pH_was_exceeded = []  # List to store indices of days where pH exceeded limits

    # Loop through each concentration in the array
    for i in range(len(concentrations_list)):
        # Calculate pH value for the concentration
        pH_value = pH_approximation(concentrations_list[i], temperature)
        # Append pH value to the list
        pH_values.append(pH_value)
        
        # Check if the pH value is within specified limits for the plant
        if plant in pH_limit_of_plants:
            if pH_limit_of_plants[plant][0] < pH_value < pH_limit_of_plants[plant][1]:
                pass  # pH within limits, no action required
            else:
                # pH exceeds limits, add index to the list
                Days_where_pH_was_exceeded.append(i)
        else:
            if 6.0 < pH_value < 7.0:
                pass  
            else:
                # pH exceeds limits, add index to the list
                Days_where_pH_was_exceeded.append(i)
    return pH_values, Days_where_pH_was_exceeded


def pH_graph(pH_values:list,Days_where_pH_was_exceeded:list,plant:str)->None:
    """"
    Generates a graph of pH values over time, with marked points where pH exceeded limits and a table of points of excess pH levels. The two figures
    are saved as png files. 
    Parameters:
        pH_values (tuple of lists of pH values over time)
        Days_where_pH_was_exceeded (list)
        plant (str)
    Returns:
        None

    """
    # Define acceptable pH range for a specific plant
    pH_range_min =pH_limit_of_plants[plant][0]
    pH_range_max =pH_limit_of_plants[plant][1]

    # Plot pH values
    plt.figure(figsize=(12, 6), dpi = 500)
    plt.plot(pH_values, marker='o', label='pH values')
    
    color = 'red'  # Color for the marked points

    for idx in Days_where_pH_was_exceeded:
        plt.plot(idx, pH_values[idx], marker='o', color=color)  # Plot each marked point with the specified color

    # Draw horizontal lines for acceptable pH range
    plt.axhline(y=pH_range_min, color='r', linestyle='--', label='Acceptable range (min)')
    plt.axhline(y=pH_range_max, color='g', linestyle='--', label='Acceptable range (max)')

    # Add legend
    plt.legend()

    # Add labels and title
    plt.xlabel('Days')
    plt.ylabel('pH')
    plt.title('pH Values Over Time')

    #Save Graph to the current directory
    current_dir = os.path.dirname(os.path.realpath(__file__))
    file_name = "graph_pH"
    file_path = os.path.join(current_dir, file_name)
    plt.savefig(file_path + ".png", format='png')  # Save as a PNG file for matplotlib

    # Initialize an empty list to hold dictionaries
    data_dict_list = []

    for i in Days_where_pH_was_exceeded:
        # Create a dictionary for each row
        row_dict = {
            'Day': i,
            'pH': pH_values[i]
        }
        # Append the dictionary to the list
        data_dict_list.append(row_dict)

    # Convert the list of dictionaries to a DataFrame
    table_of_points_of_excess = pd.DataFrame(data_dict_list)
        # Plot the table with a title
    fig, ax = plt.subplots()
    fig.patch.set_visible(False)
    ax.axis('off')
    ax.axis('tight')
    ax.table(cellText=table_of_points_of_excess.values,
             colLabels=table_of_points_of_excess.columns,
             cellLoc='center', loc='upper center')
    plt.title("Points of Excess pH Levels")
    
    #Save Graph to the current directory
    current_dir = os.path.dirname(os.path.realpath(__file__))
    file_name = "Table_pH"
    file_path = os.path.join(current_dir, file_name)
    plt.savefig(file_path + ".png", format='png')  # Save as a PNG file for matplotlib
'''
# Example of values for testing pH graph
pH_values = [6.4, 6.5, 6.8, 6.2, 6.4, 6.4, 6.0, 5.8]
Days_where_pH_was_exceeded=[6,7]
pH_graph(pH_values,Days_where_pH_was_exceeded,"tomatoes")
'''

def pH_part_of_report_generation(concentrations_list:list, temperature:float, plant:str)->None:
    """"
    Generates a pH graph and a table of points of excess pH levels by calling the generation_of_pH _list function followed 
    by the pH_graph function, the latter saving the figures as png files. 
    Parameters:
        concentrations_list (dict) 
        temperature (float)
        plant (str)
    Returns:
        None
    """
    pH_values, Days_where_pH_was_exceeded=generation_of_pH_list(concentrations_list, temperature, plant)
    pH_graph(pH_values,Days_where_pH_was_exceeded,plant)

    
    



"""Simulation of hydroponic farming."""
from .Basic_functions import *
from .Generate_Report import *
from .Solutions_Solubility import *
from .pH_graph_0 import *
from .PH_Approximation_0 import *
from .Refill_0 import *



"""This file contains tests for the functions in the Solutions_Solubility.py file."""

# Importing functions from Basic_functions.py file and Solutions_Solubility.py file
import sys
import os
sys.path.append(os.path.dirname(os.path.realpath(__file__)) + "/../src")
from hydroponics.Basic_functions import *
from hydroponics.Solutions_Solubility import *


#-----------------Test Check_solubility function--------------------------------
 
def test_valid_solubility_salt_input_bool_output():
    # Test with soluble salts
    solution = {"MnCl2" : 0.0223, "KNO3" : 0.493, "Ca(NO3)2" : 0.945}
    assert check_solubility(solution, input_type="salt", output_type="bool") == True, "check_solubility: Test valid_solubility_salt_input_bool_output failed"

def test_valid_solubility_salt_input_analysis_output():
    # Test with soluble salts
    solution = {"MnCl2" : 0.0223, "KNO3" : 0.493, "Ca(NO3)2" : 0.945}
    assert check_solubility(solution, input_type="salt", output_type="analysis") == "all salts are soluble", "check_solubility: Test valid_solubility_salt_input_analysis_output failed"

def test_valid_solubility_ion_input_update_output():
    # Test with soluble salts
    solution = {"Mn": 0.05, "Cl": 0.05, "K": 0.03, "NO3": 0.02}
    assert check_solubility(solution, input_type="ion", output_type="update") == [], "check_solubility: Test valid_solubility_ion_input_update_output failed"

def test_insoluble_solution_salt_input_bool_output():
    # Test with insoluble salts
    solution = {"MnCl2" : 0.0223, "KNO3" : 493, "Ca(NO3)2" : 0.945}
    assert check_solubility(solution, input_type="salt", output_type="bool") == False, "check_solubility: Test insoluble_solution_salt_input_bool_output failed"
    
def test_insoluble_solution_salt_input_analysis_output():
    # Test with insoluble salts
    solution = {"MnCl2" : 0.0223, "KNO3" : 493, "Ca(NO3)2" : 0.945}
    output = "The following salts precipitate: Ca(NO3)2    KNO3    "
    assert check_solubility(solution, input_type="salt", output_type="analysis") == output,"check_solubility: Test insoluble_solution_salt_input_analysis_output failed"

def test_invalid_input_type():
    # Test with invalid input type
    solution = {"NaCl": 0.1, "KBr": 0.05}
    try:
        check_solubility(solution, input_type="invalid", output_type="bool")
    except ValueError as e:
        assert str(e) == "Invalid input type. Please choose either 'salt' or 'ion'.", "check_solubility: Test invalid_input_type failed"

def test_invalid_output_type():
    # Test with invalid output type
    solution = {"NaCl": 0.1, "KBr": 0.05}
    try:
        check_solubility(solution, input_type="salt", output_type="invalid")
    except ValueError as e:
        assert str(e) == "Invalid output type. Please choose either 'bool', 'analysis', or 'update'.", "check_solubility: Test invalid_output_type failed"
        
        
        
#-----------------Test analyse_nutriments function--------------------------------
def test_ion_input():
    """
    Test with ion input
    """
    solution = {"Na+": 0.1, "Cl-": 0.1, "K+": 0.05, "NO3(-)": 0.05}
    plant = {"Na+": 0.2, "Cl-": 0.2, "K+": 0.1, "NO3(-)": 0.1}
    growth_time = 10
    volume = 1
    assert analyse_nutriments(solution, plant, growth_time, volume, input_type_solution = "ion") == [False, 5, 'Na+'], "analyse_nutriments: Test ion_input failed"

def test_salt_input():
    solution = {"NaCl":0.1, "KNO3": 0.05}
    plant = {"Na+": 0.6, "Cl-": 0.9, "K+": 0.8, "NO3(-)": 0.7}
    growth_time = 100
    volume = 2
    expected_output = [True, 100, None]
    assert analyse_nutriments(solution, plant, growth_time, volume) == expected_output, "analyse_nutriments: Test salt_input failed"

def test_negative_volume():
    solution = {"NaCl":0.1, "KNO3": 0.05}
    plant = {"Na+": 0.2, "Cl-": 0.2, "K+": 0.1, "NO3(-)": 0.1}
    growth_time = 10
    volume = -1
    try:
        analyse_nutriments(solution, plant, growth_time, volume)
    except ValueError as e:
        assert str(e) == "Volume must be positive.", "analyse_nutriments: Test negative_volume failed"
        
#-----------------Test update_sol() function--------------------------------
def test_update_sol_basic():
    # Define a simple ions_solution and plant requirements
    ions_solution = {
        'Na': 5.0,
        'K': 2.0,
        'Ca': 1.0
    }
    plant = {
        'Na': 1.0,
        'K': 2.0,
        'Ca': 0.5
    }

    # Expected outcome after one day
    expected_solution = {
        'Na': 4.0,
        'K': 0.0,
        'Ca': 0.5
    }

    result = update_sol(ions_solution, plant, 1)
    assert result == expected_solution, f"Test failed: {result} != {expected_solution}"


        
    




import time

""""
Optional: Implement the following function that will be called by the main function that is provided below.

def Concentration_Checker(frequency:float)->None:
    print("Monitoring of Concentration started at:", time.time())

# Converts frequency in number of times per day to per second (e.g., 2 times per second)
frequency =frequency/(24*60*60) # calls per second

# Calculate the time interval between function calls
time_interval = 1 / frequency # seconds

while True:
    Refill_of_container(concentration_of_ions, acceptable_maximum_percentage_deviation_from_optimal, optimal_ion_concentrations)
    time.sleep(time_interval)    
"""


def Refill_of_container(
    concentration_of_ions: dict[str, float], 
    optimal_ion_concentrations: dict[str, float], 
    *args: float
) -> dict[str, float]:
    """
    Refill the container with ions that are below the optimal concentration.
    
    This function optionally checks if the concentration of ions in the solution is within an acceptable 
    range of the optimal concentration. If specified, it skips ions that are within the acceptable deviation 
    and refills only those that are below the optimal concentration.

    Parameters:
        concentration_of_ions (dict[str, float]): A dictionary containing the concentration of ions in the solution.
        optimal_ion_concentrations (dict[str, float]): A dictionary containing the optimal concentration of ions in the solution.
        *args (float): A list of acceptable maximum percentage deviations from the optimal concentration of ions in the solution.

    Returns:
        dict[str, float]: A dictionary containing the ions and the quantities to add to the solution.
    """
    # This part calculates the quantities of ions to add to the solution
    Ion_quantities_to_add={}
    for ion, concentration in concentration_of_ions.items():
            concentration_to_add=optimal_ion_concentrations[ion]-concentration_of_ions[ion]
            Ion_quantities_to_add[ion]= concentration_to_add
    return Ion_quantities_to_add

            
    